import unittest

from siibra import volumesrc

class TestVolumeSrc(unittest.TestCase):

    def test_volume_src_init(self):
        id='test_id'
        name='test_name'
        v_type='ng_precomputed'
        url='http://localhost/test'
        volume_src = volumesrc.VolumeSrc(id, name, url)
        self.assertIsNotNone(volume_src)
        self.assertEqual(volume_src.get_url(), url)

    def test_volume_from_valid_json(self):
        v_json = {
            '@id': 'json_id',
            '@type': 'fzj/tmp/volume_type/v0.0.1',
            'name': 'json_name',
            'volume_type': 'nii',
            'url':'http://localhost/test'
        }
        output = volumesrc.from_json(v_json)
        self.assertIsInstance(output, volumesrc.VolumeSrc)

    def test_volume_from_invalid_json(self):
        v_invalid_json = {
            '@id': 'json_id',
            '@type': 'fzj/tmp/not_volume_type/v0.0.1',
            'name': 'json_name',
            'volume_type': 'json_volume_type',
            'url':'http://localhost/test'
        }
        with self.assertRaises(NotImplementedError):
            output = volumesrc.from_json(v_invalid_json)

if __name__ == "__main__":
    unittest.main()



# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from . import logger,__version__
from .commons import create_key
from .retrieval import CACHEDIR
from gitlab import Gitlab
from gitlab.exceptions import GitlabError
from requests.exceptions import ConnectionError
from tempfile import mkstemp
from tqdm import tqdm
import os
from collections import defaultdict
import re
from datetime import datetime
from glob import glob

# Until openminds is fully supported, 
# we store atlas configurations in a gitlab repo.
# We tag the configuration with each release
GITLAB_PROJECT_TAG=os.getenv("SIIBRA_CONFIG_GITLAB_PROJECT_TAG", "siibra-{}".format(__version__))

class ConfigurationRegistry:
    """
    Registers atlas configurations from siibra configuration files 
    managed in a separately maintained gitlab configuration repository.
    Each json file is converted to a specific object class based on the object construction
    function provided as constructor parameter.  
    
    The target class is determined from the folder of the configuration 
    repository (atlases, parcellations, spaces).    
    """

    GITLAB_CONFIGURATION_REPOSITORIES=[{
        'SERVER': 'https://jugit.fz-juelich.de',
        'PROJECT_ID': 3484,
    }, {
        'SERVER': 'https://gitlab.ebrains.eu',
        'PROJECT_ID': 93,
    }]
    
    logger.debug(f"Configuration: {GITLAB_PROJECT_TAG}")

    def __load_config(self,config_folder):
        """
        Find, load and cache siibra configuration files from the separately maintained gitlab configuration repository.
        """

        # try to connect to a configuration server
        want_branch_commit = None
        for gitlab_config in self.GITLAB_CONFIGURATION_REPOSITORIES:
            try:
                GITLAB_SERVER=gitlab_config.get('SERVER')
                GITLAB_PROJECT_ID=gitlab_config.get('PROJECT_ID')
                if GITLAB_SERVER is None or GITLAB_PROJECT_ID is None:
                    raise ValueError('Both SERVER and PROJECT_ID are required to determine a siibra configuration repository')
                logger.debug(f'Attempting to connect to {GITLAB_SERVER}')
                project = Gitlab(gitlab_config['SERVER'], timeout=10).projects.get(GITLAB_PROJECT_ID)
                matched_branches = list(filter(lambda b:b.name==GITLAB_PROJECT_TAG,project.branches.list()))
                if len(matched_branches)>0:
                    want_branch_commit = matched_branches[0].commit
                repository_reached = True
                break
            except (ConnectionError,GitlabError):
                # Gitlab server down. Try the next one.
                logger.debug(f'Gitlab server at {GITLAB_SERVER} unreachable. Trying next mirror.')
            except ValueError:
                logger.warn('Gitlab configuration malformed')
        else:
            # will not be reached if the for loop is broken.
            repository_reached = False
            logger.debug(f'No access to any repository with configurations for {config_folder}.')

        # Decide wether to access a possibly cached configuration
        cachefile = None
        basename = f"{config_folder}_{GITLAB_PROJECT_TAG}"
        if want_branch_commit is not None:
            sid = want_branch_commit['short_id']
            tstamp = datetime.fromisoformat(want_branch_commit['created_at']).strftime('%Y%m%d%H%M%S')
            basename += f"_commit{sid}_{tstamp}"
        if repository_reached:
            # we did connect to a configuration repository above, 
            # so we can rely on our information on the project tag.
            cachefile = os.path.join(CACHEDIR,f"config_{basename}.json")
        else:
            # We seem to be offline, so we take the best we can get from the local cache.
            cachefiles_available = glob(os.path.join(CACHEDIR,f"config_{basename}*.json"))
            if len(cachefiles_available)>0:
                logger.debug(f"Cannot connect to repository. Looking for most recently cached configuration '{GITLAB_PROJECT_TAG}' for {config_folder}.")
                if len(cachefiles_available)==1:
                    # this is either the unique cached version of a configuration tag, 
                    # or the single cached version of a branch.
                    cachefile = cachefiles_available[0]
                else:
                    # We have multiple commits cached from the same branch.
                    # Choose the one with newest timestamp.
                    get_tstamp = lambda fn: fn.replace('.json','').split('_')[-1]
                    sorted_cachefiles = sorted(cachefiles_available,key=get_tstamp)
                    cachefile = sorted_cachefiles[-1]

        if cachefile is not None and os.path.isfile(cachefile):
            # we do have a cache! read and return
            logger.info(f"Loading cached configuration '{GITLAB_PROJECT_TAG}' for {config_folder}")
            with open(cachefile,'r') as f:
                return json.load(f)

        # No cached configuration found. 
        if not repository_reached:
            raise RuntimeError(f"Cannot initialize atlases: No cached configuration data for '{GITLAB_PROJECT_TAG}'', and no access to any of the configuration repositories either.")

        # Load configuration from  repository.
        config = {}
        for node in project.repository_tree(ref=GITLAB_PROJECT_TAG):
            if node['type']!='tree' or node['name']!=config_folder:
                continue
            files = list(filter(
                lambda v: v['type']=='blob' and v['name'].endswith('.json'),
                project.repository_tree(path=config_folder,ref=GITLAB_PROJECT_TAG,all=True) ))
            msg=f"Retrieving configuration '{GITLAB_PROJECT_TAG}' for {config_folder:15.15}"
            for configfile in tqdm(files,total=len(files),desc=msg,unit=" files"):
                # retrieve the config file contents and store in cache.
                fname = configfile['name']
                config[fname] = os.path.join(CACHEDIR,f"{basename}_{fname}")
                p = project.files.get(file_path=config_folder+"/"+fname, ref=GITLAB_PROJECT_TAG)
                with open(config[fname],'wb') as f:
                    f.write(p.decode())

        logger.debug("Creating cachefile",cachefile)
        with open(cachefile,'w') as f:
            json.dump(config,f,indent='\t')

        return config

    def __init__(self,config_subfolder,cls):
        """
        Populate a new registry from the json files in the package path, using
        the "from_json" function of the provided class as hook function.
        """
        logger.debug("Initializing registry of type {} for {}".format(
            cls,config_subfolder))

        config = self.__load_config(config_subfolder)
        
        self.items = []
        self.by_key = {}
        self.by_id = {}
        self.by_name = {}
        self.cls = cls
        loglevel = logger.getEffectiveLevel()
        logger.setLevel("ERROR") # be quiet when initializing the object
        for fname,cachefile in config.items():
            with open(cachefile,'r') as f:
                obj = json.load(f, object_hook=cls.from_json)
            if not isinstance(obj,cls):
                raise RuntimeError(f'Could not generate object of type {cls} from configuration {fname}')
            key = create_key(str(obj))
            identifier = obj.id
            self.items.append(obj)
            self.by_key[key] = len(self.items)-1
            self.by_id[identifier] = len(self.items)-1
            self.by_name[obj.name] = len(self.items)-1
        logger.setLevel(loglevel)

        
    def __getitem__(self,index):
        """
        Item access is implemented either by sequential index, key or id.
        """
        if isinstance(index,int) and index<len(self.items):
            return self.items[index]
        elif isinstance(index,self.cls) and (index in self.items):
            # index is itself already an object of this registry - forward
            return index
        elif index in self.by_key:
            return self.items[self.by_key[index]]
        elif index in self.by_id:
            return self.items[self.by_id[index]]
        elif isinstance(index,str):
            # if a string is specified, we check if each word is matched in a space name
            words = [w for w in re.split('[ -]',index)]
            squeezedname = lambda item: item.name.lower().replace(" ","")
            matches = [i for i in self.items
                        if all(w.lower() in squeezedname(i) for w in words)
                        or index.replace(" ","") in squeezedname(i)]
            if len(matches)==1:
                return matches[0]
            elif len(matches)>1:
                # see if the matches are from the same collection - then we disambiguate to the newest version
                try:
                    collections = {m.version.collection for m in matches}
                    if len(collections)==1:
                        return sorted(matches,key=lambda m:m.version,reverse=True)[0]
                except Exception as e:
                    pass
                namelist = ", ".join(m.name for m in matches)
                logger.warning(f"Specification '{index}' yielded {len(matches)} different matches: {namelist}")
        raise IndexError(f"Cannot identify item '{index}' in {self.cls.__name__} registry")

    def __len__(self):
        return len(self.items)

    def __dir__(self):
        return list(self.by_key.keys()) + list(self.by_id.keys())

    def __str__(self):
        return "\n".join([i.key for i in self.items])

    def __contains__(self,index):
        return index in self.__dir__()

    def __getattr__(self,name):
        if name in self.by_key.keys():
            return self.items[self.by_key[name]]
        else:
            raise AttributeError("No such attribute: {}".format(name))




# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import numpy as np
import warnings
from gitlab import Gitlab

from .. import logger
from ..commons import ParcellationIndex
from ..region import Region
from .feature import RegionalFeature,GlobalFeature
from .extractor import FeatureExtractor
from ..termplot import FontStyles as style
from .. import termplot,parcellations

class ConnectivityProfile(RegionalFeature):

    show_as_log = True

    def __init__(self, region, profile, column_names, src_name, src_info, src_file, parcellation, kg_schema, kg_id):
        RegionalFeature.__init__(self,region)
        self.profile = profile
        self.src_name = src_name
        self.src_info = src_info
        self.src_file = src_file
        self.column_names = column_names
        self.parcellation = parcellation
        self.kg_schema = kg_schema
        self.kg_id = kg_id
        self.globalrange = None

    def __str__(self):
        """
        Returns a multiline barplot of the sorted profiles in log scale.
        Set ConnectivityProfile.show_as_log to False to see original values
        """
        if ConnectivityProfile.show_as_log:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                profile = np.log(self.profile)
                vrange = np.log(self.globalrange)
            profile[profile==-np.inf] = 0
            vrange[vrange==-np.inf] = 0
        else:
            profile = self.profile
            vrange = self.globalrange
        calib = termplot.calibrate(vrange, self.column_names)
        return "\n".join([style.BOLD+'Connectivity profile of "{}"'.format(self.region)+style.END] 
                + [ termplot.format_row(self.column_names[i] if i in self.column_names else 'Undecoded region',profile[i],calib)
                    for i in np.argsort(profile)
                    if self.profile[i]>0
                    ])


    def decode(self,parcellation,minstrength=0,force=True):
        """
        Decode the profile into a list of connections strengths to real regions
        that match the given parcellation. 
        If a column name for the profile cannot be decoded, a dummy region is
        returned if force==True, otherwise we fail.
        """
        def decoderegion(parcellation,regionname,force):
            try:
                return parcellation.decode_region(regionname)
            except ValueError as e:
                logger.warning(f'Region name "{regionname}" cannot be decoded by parcellation {parcellation.key}, returning dummy region.')
                if force:
                    return Region(regionname, parcellation, ParcellationIndex(None,None))
                raise e

        decoded = ( (strength,decoderegion(parcellation,regionname,force))
                for strength,regionname in zip(self.profile,self.column_names.values())
                if strength>minstrength)
        return sorted(decoded,key=lambda q:q[0],reverse=True)

class ConnectivityProfileExtractor(FeatureExtractor):

    _FEATURETYPE = ConnectivityProfile
    __jsons = None

    def __init__(self,atlas):

        FeatureExtractor.__init__(self,atlas)
        self.profiles = []        
        if self.__class__.__jsons is None:
            self.__load_jsons()
        
        self.load_profiles()

        for profile in self.profiles:
            self.register(profile)

    def __load_jsons(self):

        project = Gitlab('https://jugit.fz-juelich.de').projects.get(3009)
        jsonfiles = [f['name'] 
                for f in project.repository_tree(all=True)
                if f['type']=='blob' 
                and f['name'].endswith('json')]
        self.__class__.__jsons=[]
        for jsonfile in jsonfiles:
            f = project.files.get(file_path=jsonfile, ref='master')
            data = json.loads(f.decode())
            self.__class__.__jsons.append({
                'data':data,
                'jsonfile': jsonfile})
        
    def load_profiles(self):
        if self.__class__.__jsons is None:
            raise RuntimeError('must call __load_jsons before calling load_profiles !')   

        minval = maxval = 0
        for obj in self.__class__.__jsons: 
            data=obj['data']
            jsonfile=obj['jsonfile']

            src_name = data['name']
            src_info = data['description']
            kg_schema = data['kgschema'] if 'kgschema' in data.keys() else ''
            kg_id = data['kgId'] if 'kgId' in data.keys() else ''
            src_file = jsonfile
            try:
                parcellation = parcellations[data['parcellation id']]
                if parcellation!=self.parcellation:
                    logger.debug(f'parcellation with id {parcellation.id} does not match selected parcellation with id {self.parcellation.id}, ignoring {jsonfile}')
                    continue
            except IndexError as e:
                # cannot find parcellation from parcellation id
                # Log and continue
                logger.warning(f'{e}, ignoring {jsonfile} ')
                continue
            column_names = data['data']['field names']
            valid_regions = {}
            for i,name in enumerate(column_names):
                try:
                    region = parcellation.decode_region(name)
                except ValueError:
                    logger.debug(f'Cannot decode {name} at index {i}')
                    continue
                valid_regions[i] = region
            if len(valid_regions)<len(column_names):
                logger.warning(f'{len(valid_regions)} of {len(column_names)} columns in connectivity dataset pointed to valid regions.')
            for i,region in valid_regions.items():
                regionname = column_names[i]
                profile = data['data']['profiles'][regionname]
                if max(profile)>maxval:
                    maxval = max(profile)
                if min(profile)>minval:
                    minval = min(profile)
                self.profiles.append( ConnectivityProfile(
                    region, profile, 
                    {i: r.name for i,r in valid_regions.items()},
                    src_name, src_info, src_file,
                    parcellation,
                    kg_schema, kg_id ) )

        for profile in self.profiles :
            profile.globalrange = (minval,maxval)


class ConnectivityMatrix(GlobalFeature):

    def __init__(self, parcellation, matrix, column_names, src_name, src_info ):
        GlobalFeature.__init__(self,parcellation)
        self.matrix = matrix
        self.column_names = column_names
        self.src_name = src_name
        self.src_info = src_info

    def __str__(self):
        # TODO implement a reasonable display of the matrix
        return "Connectivity matrix for {}".format(
                self.parcellation)


class ConnectivityMatrixExtractor(FeatureExtractor):

    _FEATURETYPE = ConnectivityMatrix
    __features = None

    def __init__(self,atlas):

        FeatureExtractor.__init__(self,atlas)
        if self.__class__.__features is None:
            self._load_features()
        for feature in self.__class__.__features:
            self.register(feature)        

    def _load_features(self):

        project = Gitlab('https://jugit.fz-juelich.de').projects.get(3009)
        jsonfiles = [f['name'] 
                for f in project.repository_tree() 
                if f['type']=='blob' 
                and f['name'].endswith('json')]
        self.__class__.__features = []
        for jsonfile in jsonfiles: 
            f = project.files.get(file_path=jsonfile, ref='master')
            profiles = []
            data = json.loads(f.decode())
            src_name = data['name']
            src_info  = data['description']
            parcellation = parcellations[data['parcellation id']]
            if parcellation!=self.parcellation:
                continue

            # determine the valid brain regions defined in the file,
            # as well as their indices
            column_names = data['data']['field names']
            valid_regions = {}
            for i,name in enumerate(column_names):
                try:
                    region = parcellation.decode_region(name)
                except ValueError:
                    continue
                valid_regions[i] = region
            if len(valid_regions)<len(column_names):
                logger.info(f'{len(valid_regions)} of {len(column_names)} columns in connectivity dataset point to valid regions.')

            for i,region in valid_regions.items():
                regionname = column_names[i]
                profile = [data['data']['profiles'][regionname][i] 
                        for i in valid_regions.keys()]
                profiles.append(profile)
            matrix = np.array(profiles)
            assert(all(N==len(valid_regions) for N in matrix.shape))
            self.__class__.__features.append(
                ConnectivityMatrix(
                    parcellation, matrix, column_names, src_name, src_info ))

if __name__ == '__main__':

    profileextractor = ConnectivityProfileExtractor()
    matrixextractor = ConnectivityMatrixExtractor()




import os
import unittest
from siibra.atlas import REGISTRY
from siibra import atlas,parcellations,modalities
from test.get_token import get_token

token = get_token()
os.environ['HBP_AUTH_TOKEN'] = token["access_token"]

class TestAtlas(unittest.TestCase):

    JSON_ATLAS_ID = "json_1337"
    JSON_ATLAS_NAME = "JSON Human Atlas"
    atlas_as_json = {
        "@id": JSON_ATLAS_ID,
        "name": JSON_ATLAS_NAME,
        "order": 1,
        "spaces": [
            "minds/core/referencespace/v1.0.0/dafcffc5-4826-4bf1-8ff6-46b8a31ff8e2",
        ],
        "parcellations": [
            "minds/core/parcellationatlas/v1.0.0/94c1125b-b87e-45e4-901c-00daee7f2579-290",
            "minds/core/parcellationatlas/v1.0.0/94c1125b-b87e-45e4-901c-00daee7f2579",
        ]
    }
    @classmethod
    def tearDownClass(cls):
        cls.atlas.select_parcellation('2.9')
        cls.atlas.clear_selection()

    @classmethod
    def setUpClass(cls):
        cls.atlas = REGISTRY.MULTILEVEL_HUMAN_ATLAS
        cls.ATLAS_NAME = 'Multilevel Human Atlas'

    def test_atlas_init(self):
        a = atlas.Atlas('juelich/iav/atlas/v1.0.0/1', self.ATLAS_NAME)
        self.assertEqual(a.name, self.ATLAS_NAME)
        self.assertEqual(a.key, 'MULTILEVEL_HUMAN_ATLAS')
        self.assertEqual(a.id, 'juelich/iav/atlas/v1.0.0/1')
        
        # on init, parcellations and spaces are empty lists
        self.assertTrue(len(a.parcellations) == 0)
        self.assertTrue(len(a.spaces) == 0)
        
        self.assertIsNone(a.selected_region)
        self.assertIsNone(a.selected_parcellation)

    def test_parcellations(self):
        parcellations = self.atlas.parcellations
        self.assertTrue(len(parcellations) >= 11)
        self.atlas.select_parcellation(self.atlas.parcellations[0])
        self.assertIsNotNone(self.atlas.selected_parcellation)

    def test_spaces(self):
        spaces = self.atlas.spaces
        self.assertTrue(len(spaces) == 4)

    def test_to_string(self):
        atlas_str = str(self.atlas)
        self.assertEqual(atlas_str, self.ATLAS_NAME)

    def test_from_json(self):
        json_atlas = atlas.Atlas.from_json(self.atlas_as_json)
        
        self.assertTrue(type(json_atlas) is atlas.Atlas)

        self.assertEqual(json_atlas.name, self.JSON_ATLAS_NAME)
        self.assertEqual(json_atlas.id, self.JSON_ATLAS_ID)
        self.assertTrue(len(json_atlas.spaces) == 1)
        self.assertTrue(len(json_atlas.parcellations) == 2)

    def test_from_json_with_invalid_id(self):
        invalid_atlas_json = {
            "@id": "foo",
            "bar": "bar",
            "order": 1,
        }
        json_atlas = atlas.Atlas.from_json(invalid_atlas_json)
        self.assertTrue(type(json_atlas) is not atlas.Atlas)
        self.assertEqual(json_atlas, invalid_atlas_json)

    def test_from_json_with_invalid_json(self):
        # Error handling for wrong json input
        pass

    def test_select_parcellation(self):
        selected_parcellation = self.atlas.selected_parcellation
        self.assertEqual(selected_parcellation, 'Julich-Brain Cytoarchitectonic Maps 2.9')

        new_parcellation = self.atlas.parcellations[2]
        self.atlas.select_parcellation(new_parcellation)

        self.assertEqual(self.atlas.selected_parcellation, new_parcellation)
        self.assertNotEqual(self.atlas.selected_parcellation, 'Julich-Brain Cytoarchitectonic Maps 2.9')

    def test_get_map(self):
        # test downloading map
        pass

    def test_get_mask(self):
        # test the binary mask for space
        pass

    def test_get_template(self):
        # test downloading template
        pass

    def test_select_region_from_instance(self):
        pass

    def test_select_region_from_key(self):
        pass

    def test_select_region_same_as_previous(self):
        pass

    def test_clear_selection(self):
        pass

    def test_region_selected(self):
        pass

    def test_coordinate_selected(self):
        pass

    def test_get_features(self):
        # more than one test needed
        # wrong modality, no region selected
        # GeneExpression
        pass

    def test_get_features_ebrains_features(self):

        self.atlas.select_region('hoc1 left')
        features=self.atlas.get_features(modalities.EbrainsRegionalDataset)
        assert(len(features) > 0)

    def test_get_features_connectivity_profile_filter_by_sel_parc(self):
        
        expected_p_id='minds/core/parcellationatlas/v1.0.0/94c1125b-b87e-45e4-901c-00daee7f2579-290'
        self.atlas.select_parcellation(parcellations[expected_p_id])
        self.atlas.select_region('hoc1 left')
        conns=self.atlas.get_features(modalities.ConnectivityProfile)
        # TODO enable once conn data for 2.9 is added
        # assert(len(conns)>0)
        # assert(all([conn.parcellation.id == expected_p_id for conn in conns]))

    def test_regionsprops(self):
        pass

    def test_regionprops_summarize(self):
        pass


if __name__ == "__main__":
    unittest.main()



# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import logger,spaces
from .commons import OriginDataInfo, create_key, Glossary, MapType, ParcellationIndex,HasOriginDataInfo
from .space import Space
from .retrieval import cached_get
from . import volumesrc
import numpy as np
from skimage import measure
from nibabel.affines import apply_affine
import nibabel as nib
from memoization import cached
import re
import anytree
from typing import Union
from scipy.spatial.qhull import QhullError
import json

REMOVE_FROM_NAME=['hemisphere','-',
    # region string used in receptor features sometimes contains both/Both keywords
    # when they are present, the regions cannot be parsed properly
    'both', 'Both']

def _clear_name(name):
    for word in REMOVE_FROM_NAME:
        name = re.sub(r' *'+word,'',name)
    return name

RPROPS_REPO = "https://jugit.fz-juelich.de/t.dickscheid/brainscapes-datafeatures"
RPROPS_BRANCH = "master"
RPROPS_URL_SCHEME = f"{RPROPS_REPO}/-/raw/{RPROPS_BRANCH}/spatialprops/{{parckey}}-{{spacekey}}-spatialprops.json"

class Region(anytree.NodeMixin, HasOriginDataInfo):
    """
    Representation of a region with name and more optional attributes
    """

    def __init__(self, name, parcellation, index:ParcellationIndex, attrs={}, parent=None, children=None, volume_src={}):
        """
        Constructs a new region object.

        Parameters
        ----------
        name : str
            Human-readable name of the rgion
        parcellation : Parcellation
            the parcellation object that this region belongs to
        parcellaton : int
            the integer label index used to mark the region in a labelled brain volume
        index : ParcellationIndex
            the integer label index used to specify one of muliple available maps, if any (otherwise None)
        attrs : dict
            A dictionary of arbitrary additional information
        parent : Region
            Parent of this region, if any
        volume_src : Dict of VolumeSrc
            VolumeSrc objects indexed by (Space,MapType), representing available image datasets for this region map.
        """
        HasOriginDataInfo.__init__(self)
        self.name = _clear_name(name)
        self.key = create_key(name)
        self.parcellation = parcellation
        self.index = index
        self.attrs = attrs
        self.parent = parent
        self.volume_src = volume_src
        child_has_mapindex = False
        if children:
            self.children = children
            for c in self.children:
                c.parent = self
                c.parcellation = self.parcellation
                if c.index.map is not None:
                    child_has_mapindex=True

        if (self.index.map is None) and (not child_has_mapindex):
            self.index.map=0

    @staticmethod
    def copy(other):
        """
        copy contructor must detach the parent in to avoid problems with
        the Anytree implementation.
        """
        # create an isolated object, detached from the other's tree
        region = Region(other.name, other.parcellation, other.index, other.attrs)
        region.origin_datainfos=other.origin_datainfos

        # Build the new subtree recursively
        region.children = tuple(Region.copy(c) for c in other.children)
        for c in region.children:
            c.parent = region

        return region

    @property
    def labels(self):
        return {r.index.label for r in self if r.index.label is not None}

    @property
    def names(self):
        return Glossary([r.key for r in self])

    def __eq__(self,other):
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        """
        Identify each region by its parcellation and region key.
        """
        return hash(self.parcellation.key+self.key)

    def has_parent(self,parent):
        return parent in [a for a in self.ancestors]

    def __getattr__(self,name):
        if name in self.attrs.keys():
            return self.attrs[name]
        else:
            raise AttributeError("No such attribute: {}".format(name))

    def includes(self, region):
        """
        Determine wether this regiontree includes the given region.
        """
        return region==self or region in self.descendants

    def find(self,regionspec,select_uppermost=False):
        """
        Find regions that match the given region specification in the subtree
        headed by this region. 

        Parameters
        ----------
        regionspec : any of 
            - a string with a possibly inexact name, which is matched both
              against the name and the identifier key, 
            - an integer, which is interpreted as a labelindex,
            - a full ParcellationIndex
            - a region object
        select_uppermost : Boolean
            If true, only the uppermost matches in the region hierarchy are
            returned (otherwise all siblings as well if they match the name)

        Yield
        -----
        list of matching regions
        """
        if isinstance(regionspec,str) and regionspec in self.names:
            # key is given, this gives us an exact region
            return [anytree.search.find_by_attr(self,regionspec,name="key")]
        result = list(set(anytree.search.findall(self,
                lambda node: node.matches(regionspec))))
        if len(result)>1 and select_uppermost:
            all_results = result
            mindepth = min([r.depth for r in result])
            result = [r for r in all_results if r.depth==mindepth]
            if len(result)<len(all_results):
                logger.debug("Returning only {} parent nodes of in total {} matching regions for spec '{}'.".format(
                    len(result), len(all_results), regionspec))

        if isinstance(result,Region):
            return [result]
        elif result is None:
            return []
        else:
            return list(result)

    @cached
    def matches(self,regionspec):
        """ 
        Checks wether this region matches the given region specification. 

        Parameters
        ---------

        regionspec : any of 
            - a string with a possibly inexact name, which is matched both
              against the name and the identifier key, 
            - an integer, which is interpreted as a labelindex,
            - a full ParcellationIndex
            - a region object

        Yield
        -----
        True or False
        """
        splitstr = lambda s : [w for w in re.split('[^a-zA-Z0-9.]', s) 
                if len(w)>0]
        if isinstance(regionspec,Region):
            return self==regionspec
        elif isinstance(regionspec,int):
            # argument is int - a labelindex is expected
            return self.index.label==regionspec
        elif isinstance(regionspec,ParcellationIndex):
            return self.index==regionspec          
        elif isinstance(regionspec,str):
            # string is given, perform some lazy string matching
            words = splitstr(self.name.lower())
            name_matches = any([
                    regionspec==self.key,
                    regionspec==self.name,
                    all([w.lower() in words 
                        for w in splitstr(_clear_name(regionspec))])
                    ]) 
            return name_matches
        else:
            raise TypeError(f"Cannot interpret region specification of type '{type(regionspec)}'")

    @cached
    def build_mask(self,space : Space, resolution_mm=None ):
        """
        Returns a binary mask where nonzero values denote
        voxels corresponding to the region.

        NOTE: This is sensitive to the `continuous_map_threshold` attribute of
        the parent parcellation. If set, thresholded continuous maps will be
        preferred over labelled masks when a continuous regional map is available.

        Parameters
        ----------
        space : Space
            The desired template space.
        resolution_mm : float or None (Default: None)
            Request the template at a particular physical resolution in mm. If None,
            the native resolution is used.
            Currently, this only works for the BigBrain volume.
        """
        if not self.parcellation.supports_space(space):
            logger.error('Region "{}" does not provide a map for space "{}"'.format(
                str(self), str(space) ))

        mask = affine = None 

        if self.parcellation.continuous_map_threshold is not None:

            T = self.parcellation.continuous_map_threshold 
            if self.has_regional_map(space,MapType.CONTINUOUS):
                logger.info(f"Computing mask for {self.name} by thresholding the continuous regional map at {T}.")
                pmap = self.volume_src[space,MapType.CONTINUOUS].fetch(resolution_mm=resolution_mm)
            else:
                logger.info(f"Extracting mask for {self.name} from continuous map volume of {self.parcellation.name}.")
                pmap = self.parcellation.get_map(space,maptype=MapType.CONTINUOUS).extract_regionmap(self,resolution_mm=resolution_mm)
            if pmap is not None:
                mask = (np.asanyarray(pmap.dataobj)>T).astype('uint8')
                affine = pmap.affine

        elif self.has_regional_map(space,MapType.LABELLED):
            logger.info(f"Extracting mask for {self.name} from regional labelmap.")
            labelimg = self.volume_src[space,MapType.LABELLED].fetch(resolution_mm=resolution_mm)
            mask = labelimg.dataobj
            affine = labelimg.affine
        
        else:
            logger.info(f"Extracting mask for {self.name} from parcellation volume of {self.parcellation.name}.")
            labelmap = self.parcellation.get_map(space,maptype=MapType.LABELLED).fetchall(resolution_mm=resolution_mm)
            for mapindex,img in enumerate(labelmap):
                if self.index.map is not None:
                    if self.index.map!=mapindex:
                        continue
                if mask is None:
                    mask = np.zeros(img.dataobj.shape,dtype='uint8')
                    affine = img.affine
                # Note: self.labels holds all labelindices of a region and its descendants
                mask[np.isin(img.dataobj,list(self.labels))]=1

        if mask is None:
            raise RuntimeError(f"Could not compute mask for {self.region.name} in {space.name}.")
        else:
            return nib.Nifti1Image(dataobj=mask,affine=affine)

    def has_regional_map(self,space,maptype : MapType):
        """
        Tests wether a specific map is available for this region.

        Parameters
        ----------
        space : Space 
            Template space 
        maptype : MapType
            Type of map (e.g. continuous, labelled - see commons.MapType)
        """
        return (space,maptype) in self.volume_src

    @cached
    def get_regional_map(self,space:Space,maptype:Union[str,MapType]):
        """
        Retrieves and returns a specific map of this region, if available
        (otherwise None). This is typically a probability or otherwise
        continuous map, as opposed to the standard label mask from the discrete
        parcellation.

        Parameters
        ----------
        space : Space, or str
            Specifier for the template space 
        maptype : MapType
            Type of map (e.g. continuous, labelled - see commons.MapType)
        """
        if isinstance(maptype,str):
            maptype = MapType[maptype.upper()]
        try:
            spaceobj = spaces[space]
        except IndexError:
            logger.error(f"Cannot resolve space specification '{space}'.")
            return None
        if not self.has_regional_map(spaceobj,maptype):
            logger.warning(f"{self.parcellation.name} has no regional map for {self} in {spaceobj.name}.")
            return None
        return self.volume_src[spaceobj,maptype]

    def __getitem__(self, labelindex):
        """
        Given an integer label index, return the corresponding region.
        If multiple matches are found, return the unique parent if possible,
        otherwise create an artificial parent node.
        Otherwise, return None

        Parameters
        ----------

        regionlabel: int
            label index of the desired region.

        Return
        ------
        Region object
        """
        if not isinstance(labelindex,int):
            raise TypeError("Index access into the regiontree expects label indices of integer type")

        # first test this head node
        if self.index.label==labelindex:
            return self 

        # Consider children, and return the one with smallest depth
        matches = list(filter(lambda x: x is not None,
            [c[labelindex] for c in self.children] ))
        if matches:
            parentmatches = [m for m in matches if m.parent not in matches]
            if len(parentmatches)==1:
                return parentmatches[0]
            else:
                # create an articicial parent region from the multiple matches
                custom_parent = Region._build_grouptree(
                        parentmatches,self.parcellation)
                assert(custom_parent.index.label==labelindex)
                logger.warn("Label index {} resolves to multiple regions. A customized region subtree is returned: {}".format(
                    labelindex, custom_parent.name))
                return custom_parent
        return None

    @staticmethod
    def _build_grouptree(regions,parcellation):
        """
        Creates an artificial subtree from a list of regions by adding a group
        parent and adding the regions as deep copies recursively.
        """
        # determine appropriate labelindex
        indices = []
        for tree in regions:
            indices.extend([r.index for r in tree])
        unique = set(indices)
        index = next(iter(unique)) if len(unique)==1 else ParcellationIndex(None,None)

        group = Region(
                "Group: "+",".join([r.name for r in regions]),
                parcellation, index, 
                children=[Region.copy(r) for r in regions])
        return group


    def __str__(self):
        return f"{self.parcellation.name}: {self.name}"

    def __repr__(self):
        return  "\n".join("%s%s" % (pre, node.name)
                for pre, _, node in anytree.RenderTree(self))

    @cached
    def spatialprops(self,space,force=False):
        """
        Returns spatial region properties for connected components of this region found by analyzing the parcellation volume in the given space.

        Parameters
        ----------
        space : Space
            the space in which the computation shall be performed
        force : Boolean (Default: False)
            spatialprops will only be computed for leave regions (without
            children), except this is set to True.
        
        Return
        ------
        List of RegionProps objects, one per connected component found in the corresponding labelled parcellation map.
        """
        propfile = RPROPS_URL_SCHEME.format(parckey=self.parcellation.key,spacekey=space.key)
        logger.debug(f'Trying to load spatial region props from {propfile}')
        try:
            D = json.loads(cached_get(propfile))
        except Exception as e:
            raise RuntimeError(f"Cannot load and parse spatial property data for {self.parcellation.name} in {space.name} (requested data URL was {propfile})")
        return [c for p in D['spatialprops'] for c in p['components'] if p['region']['name']==self.name]

    def print_tree(self):
        """
        Returns the hierarchy of all descendants of this region as a tree.
        """
        print(self.__repr__())

    def __iter__(self):
        """
        Returns an iterator that goes through all regions in this subtree
        (including this parent region)
        """
        return anytree.PreOrderIter(self)

    @staticmethod
    def from_json(jsonstr,parcellation):
        """
        Provides an object hook for the json library to construct a Region
        object from a json definition.
        """

        # first construct any child objects
        # This is important due to the bottom-up way the tree gets
        # constructed # in the Region constructor.
        children = []
        if "children" in jsonstr:
            if jsonstr["children"] is not None:
                for regiondef in jsonstr["children"]:
                    children.append(Region.from_json(regiondef,parcellation))

        # Then setup the parent object
        assert('name' in jsonstr)
        name = jsonstr['name'] 
        pindex = ParcellationIndex(
            label = jsonstr['labelIndex'] if 'labelIndex' in jsonstr else None,
            map = jsonstr['mapIndex'] if 'mapIndex' in jsonstr else None)

        # Parse the volume sources in this region definition, if any
        # TODO the json structure should be simplified, the usage type clarified. 
        # @see https://github.com/FZJ-INM1-BDA/siibra-python/issues/42 
        volume_src = {}
        key2maptype = {
                'pmap' : MapType.CONTINUOUS,
                'collect' : MapType.LABELLED
                }
        if 'volumeSrc' in jsonstr:
            for space_id,space_vsources in jsonstr['volumeSrc'].items():
                space = spaces[space_id]
                vsrc_definitions = [{**vsrc,**{'key':key}}
                        for key,vsources in space_vsources.items()
                        for vsrc in vsources ]
                if len(vsrc_definitions)>1:
                    raise NotImplementedError(f"Multiple volume sources defined for region {name} and space {space_id}. This is not yet supported by siibra.")
                if len(vsrc_definitions)==0 or vsrc_definitions[0] is None:
                    raise ValueError(f"No valid volume source found!")
                vsrc = volumesrc.from_json(vsrc_definitions[0])
                key = vsrc_definitions[0]['key']
                if key not in key2maptype:
                    raise NotImplementedError(f"'volumeSrc' field has unknown key '{key}', cannot determine MapType")
                volume_src[space,key2maptype[key]] = vsrc

        r = Region( name, parcellation, pindex, 
                attrs=jsonstr, children=children, 
                volume_src=volume_src )

        if 'originDatasets' in jsonstr:
            origin_datainfos=[OriginDataInfo.from_json(ds) for ds in jsonstr.get('originDatasets', [])]
            r.origin_datainfos=[f for f in origin_datainfos if f is not None]

        # inherit labelindex from children, if they agree
        if pindex.label is None and r.children: 
            L = [c.index.label for c in r.children]
            if (len(L)>0) and (L.count(L[0])==len(L)):
                r.index.label = L[0]

        return r


if __name__ == '__main__':

    definition = {
            'name': 'Interposed Nucleus (Cerebellum) - left hemisphere',
            'rgb': [170, 29, 10],
            'labelIndex': 251,
            'ngId': 'jubrain mni152 v18 left',
            'children': [],
            'position': [-9205882, -57128342, -32224599],
            'originDatasets': [ {
                    'kgId': '658a7f71-1b94-4f4a-8f15-726043bbb52a', 
                    'kgSchema': 'minds/core/dataset/v1.0.0', 
                    'filename': 'Interposed Nucleus (Cerebellum) [v6.2, ICBM 2009c Asymmetric, left hemisphere]'
                    }]
            }




from setuptools import setup, find_packages
import os
import re

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

def find_version():
    path_to_init=os.path.join(ROOT_DIR, 'siibra', '__init__.py')
    with open(path_to_init, 'r', encoding="utf-8") as f:
        content=f.read()
        version_match=re.search(r"^__version__=['\"](.*?)['\"]$", content, re.M)
        if version_match:
            return version_match.group(1)
        raise RuntimeError('version cannot be found!')

with open(os.path.join(ROOT_DIR,"README.md"), "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="siibra",
    version=find_version(),
    author="Big Data Analytics Group, Forschungszentrum Juelich, Institute of Neuroscience and Medicine (INM-1)",
    author_email="inm1-bda@fz-juelich.de",
    description="siibra - Software interfaces for interacting with brain atlases",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/FZJ-INM1-BDA/siibra-python",
    packages=find_packages(include=['siibra', 'siibra.*']),
    include_package_data=True,
    # packages=find_packages(include=['.']),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        'Intended Audience :: Developers',
    ],
    python_requires='>=3.6',
    install_requires=['anytree','statsmodels','nibabel','click','appdirs','scikit-image','scipy','statsmodels','requests','python-gitlab','memoization','cloud-volume','nilearn'],
)




# Copyright 2018-2020 Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sys import stdout, stdin, platform
from os import popen

if stdin.isatty() and not platform.startswith('win'):
    # running interactively
    _,TERMWIDTH = [int(v) for v in popen('stty size', 'r').read().split()]
else:
    TERMWIDTH=80

# Ascii formatting, see 
# https://stackoverflow.com/questions/8924173/how-do-i-print-bold-text-in-python
class MarkerStyles:
    THIN = u'\u2584'
    MEDIUM = u'\u2586'
    THICK = u'\u2587'
    FULL = u'\u2588'

class ColorStyles:
   WHITE = '\033[37m'
   LIGHTGRAY = '\033[97m'
   GRAY = '\033[2m'
   DARKGRAY = '\033[90m'
   ORANGE = '\033[33m'
   LIGHTBLUE = '\033[94m'
   BLUE = '\033[34m'
   GREEN = '\033[32m'
   RED = '\033[91m'
   DARKRED = '\033[1m'
   END = '\033[0m'

class FontStyles:
   BOLD = '\033[1m'
   BLACK = '\033[90m'
   RED = '\033[91m'
   UNDERLINE = '\033[4m'
   ITALIC = '\033[3m'
   END = '\033[0m'

class Cursor:
    COL0 = '\033[1000D'

def bar(length,
        colorstyle=ColorStyles.DARKGRAY,
        marker=MarkerStyles.THICK,
        shift=0):
    """ 
    Prints a vertical bar with the given color and length to the console.
    Choose markers from https://en.wikipedia.org/wiki/Block_Elements
    """
    #start = Cursor.COL0 if startleft else ""
    start = u"\033[{n}C".format(n=shift-1)
    end = ""#"\n" if close_line else ""
    return start+colorstyle+marker*length+ColorStyles.END+end
    #stdout.flush()

def calibrate(values,labels,maxbars=TERMWIDTH):
    """ calculate scale and shift for bar plot groups """
    minsum = min(float(min(values)),0)
    maxsum = max(float(max(values)),0)
    labelwidth = max([len(str(l)) for l in labels])
    barwidth = maxbars-labelwidth-14
    s = (maxsum-minsum) / float(barwidth) 
    if s==0:
        blocks = lambda x: 0
    else:
        blocks = lambda x : int( (x-minsum) / s + .5)
    negative = -min(minsum,0)
    return (blocks,barwidth,negative,labelwidth)

def format_row(label,value,calibration):
    blocks,maxbars,negative,labelwidth = calibration
    result = "{0:{w}s} {1:10.2f}".format(label,value,w=labelwidth)
    if value<0:
        result += bar(blocks(value),ColorStyles.LIGHTGRAY)
        result += bar(blocks(0)-blocks(value),ColorStyles.GRAY)
        result += bar(maxbars-blocks(0),ColorStyles.LIGHTGRAY,shift=1)
    else:
        result += bar(blocks(0),ColorStyles.LIGHTGRAY)
        result += bar(blocks(value-negative),shift=1)
        result += bar(maxbars-blocks(0)-blocks(value-negative),ColorStyles.LIGHTGRAY)
    return result+" "




import unittest
from unittest.mock import MagicMock, patch

from siibra import parcellations, ebrains, spaces, retrieval
from siibra.region import Region
from siibra.commons import MapType


class TestRegions(unittest.TestCase):

    region_name = 'Interposed Nucleus (Cerebellum) left'
    kg_id = '658a7f71-1b94-4f4a-8f15-726043bbb52a'
    parentname = 'region_parent'

    definition = {
        'name': region_name,
        'rgb': [170, 29, 10],
        'labelIndex': 251,
        'ngId': 'jubrain mni152 v18 left',
        'children': [],
        'position': [-9205882, -57128342, -32224599],
        'originDatasets': [ {
            'kgId': kg_id,
            'kgSchema': 'minds/core/dataset/v1.0.0',
            'filename': 'Interposed Nucleus (Cerebellum) [v6.2, ICBM 2009c Asymmetric, left hemisphere]'
        }],
        "volumeSrc": {
            spaces[0].id : {
                "pmap": [
                    {
                        "@type": "fzj/tmp/volume_type/v0.0.1",
                        "@id": "fzj/tmp/volume_type/v0.0.1/pmap",
                        'name': 'Probabilistic map '+region_name,
                        "volume_type": "nii",
                        "url": "https://object.cscs.ch/v1/AUTH_227176556f3c4bb38df9feea4b91200c/hbp-d000001_jubrain-cytoatlas-Area-Ch-4_pub/4.2/Ch-4_l_N10_nlin2Stdcolin27_4.2_publicP_b92bf6270f6426059d719a6ff4d46aa7.nii.gz"
                    }
                ]
            },
        }
    }

    parent_definition = {
        'name': parentname,
        'rgb': [170, 29, 10],
        'labelIndex': 251,
        'ngId': 'jubrain mni152 v18 left',
        'children': [],
        'position': [-9205882, -57128342, -32224599],
        'originDatasets': [ {
            'kgId': kg_id,
            'kgSchema': 'minds/core/dataset/v1.0.0',
            'filename': 'Interposed Nucleus (Cerebellum) [v6.2, ICBM 2009c Asymmetric, left hemisphere]'
        }],
        'volumeSrc' : {}
    }
    parent_region = None
    child_region = None

    @classmethod
    def setUpClass(cls):
        retrieval.download_file = MagicMock()
        retrieval.download_file.return_value = None

        cls.parent_region = Region.from_json(cls.parent_definition,parcellations[0])

        cls.child_region = Region.from_json(cls.definition,parcellations[0])
        cls.child_region.parent=cls.parent_region

    def test_regions_init(self):
        self.assertEqual(self.child_region.name, self.region_name)

    def test_has_no_parent(self):
        self.assertFalse(self.parent_region.has_parent(self.parent_region))

    def test_has_parent(self):
        self.assertTrue(self.child_region.has_parent(self.parent_region))

    def test_includes_region_true(self):
        self.parent_region.children = [self.child_region]
        self.assertTrue(self.parent_region.includes(self.child_region))

    def test_includes_region_false(self):
        self.parent_region.children = []
        self.assertFalse(self.parent_region.includes(self.child_region))

    def test_includes_region_self(self):
        self.assertTrue(self.parent_region.includes(self.parent_region))

    def test_find_child_region(self):
        regions = self.parent_region.find(self.region_name)
        self.assertIsNotNone(regions)
        self.assertEqual(len(regions), 1)
        self.assertEqual(next(iter(regions)), self.child_region)

    def test_find_child_no_result(self):
        regions = self.child_region.find(self.parentname)
        self.assertIsNotNone(regions)
        self.assertEqual(len(regions), 0)

    def test_matches_with_valid_string(self):
        self.assertTrue(self.child_region.matches('Interposed Nucleus'))

    def test_matches_with_invalid_string(self):
        self.assertFalse(self.child_region.matches('Area 51'))

    def test_matches_with_valid_region(self):
        self.assertTrue(self.child_region.matches(self.child_region))

    def test_matches_with_wrong_region(self):
        self.assertFalse(self.child_region.matches(self.parent_region))

    @patch('siibra.parcellationmap.ParcellationMap.fetch')
    def test_regional_map_none(self, download_mock):
        self.assertIsNone(self.parent_region.get_regional_map(spaces[0],MapType.LABELLED))
        download_mock.assert_not_called()

    @patch('siibra.parcellationmap.ParcellationMap.fetch')
    def test_get_regional_map_wrong_space(self, download_mock):
        self.child_region.attrs['maps'] = {
            'spaceId': 'map_url'
        }
        self.assertIsNone(self.child_region.get_regional_map(spaces[0],MapType.LABELLED))
        download_mock.assert_not_called()

    @patch('siibra.parcellationmap.ParcellationMap.fetch')
    def test_get_regional_map_no_filename(self, fetch_mock):
        fetch_mock.return_value = None
        with self.assertRaises(RuntimeError):
            self.child_region.get_regional_map(spaces.BIG_BRAIN,MapType.CONTINUOUS).fetch()
        fetch_mock.assert_not_called()


if __name__ == "__main__":
    unittest.main()
